"""ChromaDB implementation for vector database."""

from typing import Any, List, Tuple

import numpy as np

from chromadb.api import ClientAPI

from rago.augmented.db.base import DBBase


class ChromaDB(DBBase):
    """ChromaDB implementation for vector database."""

    def __init__(
        self,
        client: ClientAPI,
        collection_name: str = 'rago',
    ) -> None:
        """Initialize ChromaDB."""
        self.client = client
        self.collection_name = collection_name
        self._setup()

    def _setup(self) -> None:
        """Set up ChromaDB client and collection."""
        self.collection = self.client.get_or_create_collection(
            name=self.collection_name
        )
        self.index = self.collection

    def embed(self, documents: Any) -> None:
        """Embed the documents into the database."""
        if not isinstance(documents, tuple) or len(documents) != 2:
            raise ValueError(
                'documents format must be: (List[str], List[List[float]])'
            )

        documents_list: List[str] = documents[0]
        embeddings_list: List[List[float]] = documents[1]

        # Convert embeddings to numpy array
        embeddings = np.array(embeddings_list, dtype=np.float32)

        self.collection.add(
            documents=documents_list,
            embeddings=embeddings,
            ids=[str(i) for i in range(len(documents_list))],
        )

    def search(
        self, query_encoded: Any, top_k: int = 2
    ) -> Tuple[List[float], List[str]]:
        """Search a query from documents."""
        # Convert query_encoded to numpy array
        query_encoded_np = np.array([query_encoded], dtype=np.float32)

        results = self.collection.query(
            query_embeddings=query_encoded_np.tolist(),
            n_results=top_k,
        )

        # Check if keys exist before accessing them
        distances = results.get('distances', [[]])
        ids = results.get('ids', [[]])

        # Ensure distances and ids are not None before indexing
        distances_list: List[float] = distances[0] if distances else []
        ids_list: List[str] = ids[0] if ids else []

        return distances_list, ids_list



"""Tests for Rago package using OpenAI GPT."""

import os
import shutil

from functools import partial
from pathlib import Path

import pytest

from rago import Rago
from rago.augmented import OpenAIAug, SpaCyAug
from rago.extensions.cache import CacheFile
from rago.generation import OpenAIGen
from rago.retrieval import PDFPathRet

PDF_DATA_PATH = Path(__file__).parent / 'data' / 'pdf'
TMP_DIR = Path('/tmp') / 'rago'

RET_CACHE = CacheFile(target_dir=TMP_DIR / 'ret')
AUG_CACHE = CacheFile(target_dir=TMP_DIR / 'aug')
GEN_CACHE = CacheFile(target_dir=TMP_DIR / 'gen')


def clear_folder(folder: Path):
    """
    Remove all files and subdirectories inside the given folder.

    Parameters
    ----------
    folder : Path
        The folder whose contents should be deleted.
    """
    if not folder.exists():
        print(f"Folder '{folder}' does not exist.")
        return

    for item in folder.iterdir():
        if item.is_file():
            item.unlink()  # Remove file
        elif item.is_dir():
            shutil.rmtree(item)  # Remove directory and its contents


def is_directory_empty(directory: Path) -> bool:
    """Check if the directory is not empty."""
    return not os.listdir(directory)


@pytest.fixture
def api_keys(env) -> {str, str}:
    """Fixture for OpenAI API key from environment."""
    keys = {}

    openai_api_key = os.getenv('OPENAI_API_KEY')
    if not openai_api_key:
        raise EnvironmentError(
            'Please set the OPENAI_API_KEY environment variable.'
        )
    keys['OpenAIAug'] = openai_api_key
    keys['OpenAIGen'] = openai_api_key

    return keys


@pytest.mark.skip_on_ci
@pytest.mark.parametrize(
    'aug_class',
    [
        partial(OpenAIAug, top_k=3, cache=AUG_CACHE),
        partial(SpaCyAug, top_k=3, cache=AUG_CACHE),
    ],
)
def test_cache(
    animals_data: list[str], api_keys: dict[str, str], aug_class: partial
) -> None:
    """Test RAG pipeline with OpenAI's GPT."""
    api_name = aug_class.func.__name__
    aug_api_key = api_keys.get(api_name, '')
    gen_api_key = api_keys.get('OpenAIGen', '')

    for cache in [RET_CACHE, AUG_CACHE, GEN_CACHE]:
        clear_folder(cache.target_dir)

    ret = PDFPathRet(PDF_DATA_PATH / '1.pdf', cache=RET_CACHE)
    aug = aug_class(api_key=aug_api_key)
    gen = OpenAIGen(
        api_key=gen_api_key, model_name='gpt-3.5-turbo', cache=GEN_CACHE
    )

    rag = Rago(
        retrieval=ret,
        augmented=aug,
        generation=gen,
    )

    query = 'Is vitamin D effective?'
    rag.prompt(query)

    # note: we don't need to test the gen_cache
    for cache in [RET_CACHE, AUG_CACHE]:
        assert not is_directory_empty(cache.target_dir), (
            f"Cache for {cache} didn't work."
        )
        clear_folder(cache.target_dir)



"""Tests for language detection."""

import pytest

from langdetect import detect


@pytest.mark.parametrize(
    'query,expected_language',
    [
        ('Are there animals bigger than a dinosaur?', 'en'),
        ("Y a-t-il des animaux plus grands qu'un dinosaure?", 'fr'),
    ],
)
def test_language_detection(
    query: str,
    expected_language: str,
) -> None:
    """Test language detection."""
    detected_language = detect(query)
    assert detected_language == expected_language, (
        f'Expected {expected_language}, got {detected_language}'
    )



"""Rago DB package."""

from __future__ import annotations

from rago.augmented.db.base import DBBase
from rago.augmented.db.chroma import ChromaDB
from rago.augmented.db.faiss import FaissDB

__all__ = ['ChromaDB', 'DBBase', 'FaissDB']



"""Tests for Rago package for augmentation."""

from functools import partial

import pytest

from rago.augmented import (
    CohereAug,
    FireworksAug,
    OpenAIAug,
    SpaCyAug,
    TogetherAug,
)

API_MAP = {
    OpenAIAug: 'api_key_openai',
    CohereAug: 'api_key_cohere',
    FireworksAug: 'api_key_fireworks',
    TogetherAug: 'api_key_together',
}

gen_models = [
    # model 0
    partial(
        SpaCyAug,
        **dict(
            model_name='en_core_web_md',
        ),
    ),
    # model 1
    partial(
        OpenAIAug,
        **dict(
            model_name='text-embedding-3-small',
        ),
    ),
    # model 2
    partial(
        CohereAug,
    ),
    # model 3
    partial(
        FireworksAug,
    ),
    # model 4
    partial(TogetherAug, **dict(model_name='BAAI/bge-base-en-v1.5')),
]


@pytest.mark.skip_on_ci
@pytest.mark.parametrize(
    'question,expected_answer',
    [
        ('Is there any animal larger than a dinosaur?', 'Blue Whale'),
        (
            'What animal is renowned as the fastest animal on the planet?',
            'Peregrine Falcon',
        ),
        ('An animal which do pollination?', 'Honey Bee'),
    ],
)
@pytest.mark.parametrize('partial_model', gen_models)
def test_aug_spacy(
    animals_data: list[str],
    question: str,
    expected_answer: str,
    api_key_openai: str,
    api_key_cohere: str,
    api_key_gemini: str,
    api_key_fireworks: str,
    api_key_together: str,
    api_key_hugging_face: str,
    partial_model: partial,
) -> None:
    """Test RAG pipeline with SpaCy."""
    logs = {'augmented': {}}
    top_k = 2

    model_class = partial_model.func

    api_key_name: str = API_MAP.get(model_class, '')
    api_key = locals().get(api_key_name, '')

    model_args = {
        'top_k': top_k,
        'logs': logs['augmented'],
        **({'api_key': api_key} if api_key else {}),
    }

    gen_model = partial_model(**model_args)

    aug_result = gen_model.search(question, animals_data)
    assert gen_model.top_k == top_k
    assert top_k >= len(aug_result)
    assert any(
        [expected_answer.lower() in result.lower() for result in aug_result]
    )

    # check if logs have been used
    assert logs['augmented']



"""FireworksGen class for text generation using Fireworks API."""

from __future__ import annotations

from typing import cast

import instructor

from fireworks.client import Fireworks
from pydantic import BaseModel
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class FireworksGen(GenerationBase):
    """Fireworks AI generation model for text generation."""

    default_model_name: str = 'accounts/fireworks/models/llama-v3-8b-instruct'
    default_api_params = {
        'top_p': 0.9,
    }

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        model = Fireworks(api_key=self.api_key)

        self.model = (
            instructor.from_fireworks(
                client=model,
                mode=instructor.Mode.FIREWORKS_JSON,
            )
            if self.structured_output
            else model
        )

    def generate(self, query: str, context: list[str]) -> str | BaseModel:
        """Generate text using Fireworks AI's API."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )

        api_params = self.api_params or self.default_api_params

        messages = []
        if self.system_message:
            messages.append({'role': 'system', 'content': self.system_message})
        messages.append({'role': 'user', 'content': input_text})

        model_params = {
            'model': self.model_name,
            'messages': messages,
            'max_tokens': self.output_max_length,
            'temperature': self.temperature,
            **api_params,
        }

        if self.structured_output:
            model_params['response_model'] = self.structured_output
            response = self.model.chat.completions.create(**model_params)
            self.logs['model_params'] = model_params
            return cast(BaseModel, response)

        response = self.model.chat.completions.create(**model_params)
        self.logs['model_params'] = model_params
        return cast(str, response.choices[0].message.content.strip())



"""Tests for Rago package using OpenAI GPT."""

import os

import pytest

from rago import Rago
from rago.augmented import OpenAIAug
from rago.generation import OpenAIGen
from rago.retrieval import StringRet


@pytest.fixture
def api_key(env) -> str:
    """Fixture for OpenAI API key from environment."""
    api_key = os.getenv('OPENAI_API_KEY')
    if not api_key:
        raise EnvironmentError(
            'Please set the OPENAI_API_KEY environment variable.'
        )
    return api_key


@pytest.mark.skip_on_ci
def test_rag_openai_gpt_general(animals_data: list[str], api_key: str) -> None:
    """Test RAG pipeline with OpenAI's GPT."""
    logs = {
        'retrieval': {},
        'augmented': {},
        'generation': {},
    }

    temperature = 0

    ret = StringRet(animals_data, logs=logs['retrieval'])
    gen = OpenAIGen(
        api_key=api_key,
        model_name='gpt-3.5-turbo',
        logs=logs['generation'],
        temperature=temperature,
    )
    aug = OpenAIAug(api_key=api_key, top_k=3, logs=logs['augmented'])

    assert gen.temperature == temperature

    rag = Rago(
        retrieval=ret,
        augmented=aug,
        generation=gen,
    )

    query = 'Is there any animal larger than a dinosaur?'
    result = rag.prompt(query)

    assert 'blue whale' in result.lower(), (
        'Expected response to mention Blue Whale as a larger animal.'
    )

    # check if logs have been used
    assert logs['retrieval']
    assert logs['augmented']
    assert logs['generation']



"""Tests for text splitter."""

from __future__ import annotations

from rago.retrieval.text_splitter import LangChainTextSplitter


def test_langchain_text_splitter_with_separator(
    animals_data: list[str],
) -> None:
    """Test the langchain text splitter."""
    text = '/n'.join(animals_data)

    # /n can also be used as separator with langchain splitter
    max_chunk_size_original = max([len(line) for line in animals_data])

    # chunk_size is the max size of the chunk
    chunk_size = max_chunk_size_original
    chunk_overlap = 100
    splitter = LangChainTextSplitter(
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
    )

    result = splitter.split(text)

    for chunk in result:
        assert len(chunk) < chunk_size

    assert len(result) >= len(animals_data)


def test_langchain_text_splitter_without_separator(
    animals_data: list[str],
) -> None:
    """Test the langchain text splitter."""
    text = '. '.join(animals_data)

    # chunk_size is the max size of the chunk
    chunk_size = len(text) // len(animals_data)
    chunk_overlap = 100
    splitter = LangChainTextSplitter(
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
    )

    result = splitter.split(text)

    for chunk in result:
        assert len(chunk) < chunk_size

    assert len(result) >= len(animals_data)



"""Tests for Rago generation module."""

import platform

from functools import partial
from typing import cast

import pytest

from rago.generation import (
    CohereGen,
    DeepSeekGen,
    FireworksGen,
    GeminiGen,
    GroqGen,
    HuggingFaceGen,
    HuggingFaceInfGen,
    LlamaGen,
    OllamaGen,
    OllamaOpenAIGen,
    OpenAIGen,
    PhiGen,
    TogetherGen,
)

from .models import AnimalModel

# LlamaGen doesn't support temperature zero
TEMPERATURE = 0.0001
GENERATION_LOG = {'generation': {}}

IS_OS_MACOS = platform.system().lower() == 'darwin'

API_MAP = {
    GeminiGen: 'api_key_gemini',
    OpenAIGen: 'api_key_openai',
    HuggingFaceGen: 'api_key_hugging_face',
    HuggingFaceInfGen: 'api_key_hugging_face',
    LlamaGen: 'api_key_hugging_face',
    OllamaGen: '',
    OllamaOpenAIGen: '',
    CohereGen: 'api_key_cohere',
    FireworksGen: 'api_key_fireworks',
    TogetherGen: 'api_key_together',
    GroqGen: 'api_key_groq',
}

gen_models_no_ci = [
    # model 0
    partial(
        OpenAIGen,
        **dict(
            model_name='gpt-3.5-turbo',
        ),
    ),
    # model 1
    partial(
        GeminiGen,
        **dict(
            model_name='gemini-1.5-flash',
        ),
    ),
    # model 2
    partial(
        HuggingFaceGen,
    ),
    # model 3
    partial(
        LlamaGen,
        **dict(device='auto'),
    ),
    # model 4
    partial(
        CohereGen,
        **dict(
            model_name='command-r-plus-08-2024',
        ),
    ),
    # model 5
    partial(
        DeepSeekGen,
        **dict(
            device='auto',
        ),
    ),
    # model 6
    partial(
        FireworksGen,
    ),
    # model 7
    partial(
        TogetherGen,
        **dict(model_name='meta-llama/Llama-3.3-70B-Instruct-Turbo-Free'),
    ),
    # model 8
    partial(
        HuggingFaceInfGen,
    ),
    # model 9
    partial(
        GroqGen,
    ),
    # model 10
    partial(
        PhiGen,
        **dict(model_name='microsoft/phi-2'),
    ),
]


gen_models_ci = [
    # model 0
    partial(
        OllamaOpenAIGen,
    ),
    # model 1
    partial(
        OllamaGen,
    ),
]


def _generation_simple_output(
    animals_data: list[str],
    api_key_openai: str,
    api_key_cohere: str,
    api_key_fireworks: str,
    api_key_gemini: str,
    api_key_together: str,
    api_key_hugging_face: str,
    api_key_groq: str,
    partial_model: partial,
) -> bool:
    """Test RAG pipeline with model generation."""
    model_class = partial_model.func

    if IS_OS_MACOS and issubclass(
        model_class,
        (
            OllamaGen,
            OllamaOpenAIGen,
        ),
    ):
        pytest.skip(f'Skipping {model_class} on macOS due to CI failure.')

    api_key_name: str = API_MAP.get(model_class, '')
    api_key = locals().get(api_key_name, '')

    model_args = {
        'temperature': TEMPERATURE,
        'logs': GENERATION_LOG['generation'],
        **({'api_key': api_key} if api_key else {}),
    }

    expected_answer = 'peregrine falcon'
    context = [
        text for text in animals_data if expected_answer in text.lower()
    ]

    gen_model = partial_model(**model_args)

    query = 'what is the fastest bird on the earth?'
    result = gen_model.generate(query, context)

    error_message = (
        f'Expected response: `{expected_answer}`, Result: `{result}`.'
    )

    assert gen_model.temperature == TEMPERATURE
    assert expected_answer.lower() in result.lower(), error_message
    assert GENERATION_LOG['generation']

    return True


@pytest.mark.skip_on_ci
@pytest.mark.parametrize('partial_model', gen_models_no_ci)
def test_generation_simple_output(
    animals_data: list[str],
    api_key_openai: str,
    api_key_cohere: str,
    api_key_fireworks: str,
    api_key_gemini: str,
    api_key_together: str,
    api_key_hugging_face: str,
    api_key_groq: str,
    partial_model: partial,
) -> None:
    """Test RAG pipeline with model generation."""
    assert _generation_simple_output(
        animals_data,
        api_key_openai,
        api_key_cohere,
        api_key_fireworks,
        api_key_gemini,
        api_key_together,
        api_key_hugging_face,
        api_key_groq,
        partial_model,
    )


@pytest.mark.parametrize('partial_model', gen_models_ci)
def test_generation_simple_output(
    animals_data: list[str],
    api_key_openai: str,
    api_key_cohere: str,
    api_key_fireworks: str,
    api_key_gemini: str,
    api_key_together: str,
    api_key_hugging_face: str,
    api_key_groq: str,
    partial_model: partial,
) -> None:
    """Test RAG pipeline with model generation."""
    assert _generation_simple_output(
        animals_data,
        api_key_openai,
        api_key_cohere,
        api_key_fireworks,
        api_key_gemini,
        api_key_together,
        api_key_hugging_face,
        api_key_groq,
        partial_model,
    )


def _generation_structure_output(
    api_key_openai: str,
    api_key_cohere: str,
    api_key_fireworks: str,
    api_key_gemini: str,
    api_key_together: str,
    api_key_hugging_face: str,
    api_key_groq: str,
    animals_data: list[str],
    question: str,
    partial_model: partial,
    expected_answer: tuple[str],
) -> bool:
    """Test Model Generation with structure output."""
    model_class = partial_model.func

    # Skip structured output for models that don't support it
    if issubclass(
        model_class,
        (
            DeepSeekGen,
            HuggingFaceInfGen,
            HuggingFaceGen,
            LlamaGen,
            OllamaGen,
        ),
    ):
        pytest.skip(f"{model_class} doesn't support structured output.")

    if IS_OS_MACOS and issubclass(
        model_class,
        (
            OllamaGen,
            OllamaOpenAIGen,
        ),
    ):
        pytest.skip(f'Skipping {model_class} on macOS due to CI failure.')

    api_key_name: str = API_MAP.get(model_class, '')
    api_key = locals().get(api_key_name, '')

    model_args = {
        'temperature': TEMPERATURE,
        'logs': GENERATION_LOG['generation'],
        'structured_output': AnimalModel,
        **({'api_key': api_key} if api_key else {}),
    }

    gen_model = partial_model(**model_args)

    context = [
        text
        for text in animals_data
        if all(
            [expected.lower() in text.lower() for expected in expected_answer]
        )
    ]

    assert gen_model.temperature == TEMPERATURE
    result = cast(AnimalModel, gen_model.generate(question, context))

    error_message = (
        f'Expected response to mention `{expected_answer}`. '
        f'Result: `{result.name}`.'
    )

    assert result.name.lower() in expected_answer, error_message
    assert GENERATION_LOG['generation']
    return True


@pytest.mark.skip_on_ci
@pytest.mark.parametrize(
    'question,expected_answer',
    [
        ('what is the fastest bird on the earth?', ('peregrine falcon',)),
    ],
)
@pytest.mark.parametrize('partial_model', gen_models_no_ci)
def test_generation_structure_output(
    api_key_openai: str,
    api_key_cohere: str,
    api_key_fireworks: str,
    api_key_gemini: str,
    api_key_together: str,
    api_key_hugging_face: str,
    api_key_groq: str,
    animals_data: list[str],
    question: str,
    partial_model: partial,
    expected_answer: tuple[str],
) -> None:
    """Test Model Generation with structure output."""
    assert _generation_structure_output(
        api_key_openai,
        api_key_cohere,
        api_key_fireworks,
        api_key_gemini,
        api_key_together,
        api_key_hugging_face,
        api_key_groq,
        animals_data,
        question,
        partial_model,
        expected_answer,
    )


@pytest.mark.parametrize(
    'question,expected_answer',
    [
        ('what is the fastest bird on the earth?', ('peregrine falcon',)),
    ],
)
@pytest.mark.parametrize('partial_model', gen_models_ci)
def test_generation_structure_output(
    api_key_openai: str,
    api_key_cohere: str,
    api_key_fireworks: str,
    api_key_gemini: str,
    api_key_together: str,
    api_key_hugging_face: str,
    api_key_groq: str,
    animals_data: list[str],
    question: str,
    partial_model: partial,
    expected_answer: tuple[str],
) -> None:
    """Test Model Generation with structure output."""
    assert _generation_structure_output(
        api_key_openai,
        api_key_cohere,
        api_key_fireworks,
        api_key_gemini,
        api_key_together,
        api_key_hugging_face,
        api_key_groq,
        animals_data,
        question,
        partial_model,
        expected_answer,
    )



"""Llama generation module."""

from __future__ import annotations

import warnings

from copy import copy
from typing import Any

import instructor
import openai
import torch

from langdetect import detect
from ollama import Client as Ollama
from pydantic import BaseModel
from transformers import AutoModelForCausalLM, AutoTokenizer, pipeline
from typeguard import typechecked

from rago.generation.base import GenerationBase
from rago.generation.openai import OpenAIGen


@typechecked
class LlamaGen(GenerationBase):
    """Llama Generation class."""

    default_model_name: str = 'meta-llama/Llama-3.2-1B'
    default_temperature: float = 0.5
    default_output_max_length: int = 500
    default_api_params = {
        'top_p': 1.0,
        'num_return_sequences': 1,
    }

    def _validate(self) -> None:
        """Raise an error if the initial parameters are not valid."""
        if not self.model_name.startswith('meta-llama/'):
            raise Exception(
                f'The given model name {self.model_name} is not provided '
                'by meta.'
            )

        if self.structured_output:
            warnings.warn(
                'Structured output is not supported yet in '
                f'{self.__class__.__name__}.'
            )

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        self.tokenizer = AutoTokenizer.from_pretrained(
            self.model_name, token=self.api_key
        )  # type: ignore

        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_name,
            token=self.api_key,
            torch_dtype=torch.float16
            if self.device_name == 'cuda'
            else torch.float32,
        )

        self.generator = pipeline(
            'text-generation',
            model=self.model,
            tokenizer=self.tokenizer,
            device=0 if self.device_name == 'cuda' else -1,
        )

    def generate(self, query: str, context: list[str]) -> str:
        """Generate text using Llama model with language support."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )

        # Detect and set the language code for multilingual models (optional)
        language = str(detect(query)) or 'en'
        self.tokenizer.lang_code = language

        api_params = (
            self.api_params if self.api_params else self.default_api_params
        )

        # Generate the response with adjusted parameters

        model_params = dict(
            text_inputs=input_text,
            max_new_tokens=self.output_max_length,
            do_sample=True,
            temperature=self.temperature,
            eos_token_id=self.tokenizer.eos_token_id,
            **api_params,
        )
        response = self.generator(**model_params)

        self.logs['model_params'] = model_params

        # Extract and return the answer only
        answer = str(response[0].get('generated_text', ''))
        # Strip off any redundant text after the answer itself
        return answer.split('Answer:')[-1].strip()


@typechecked
class OllamaGen(GenerationBase):
    """Ollama Generation class for local inference via ollama-python."""

    default_model_name = 'llama3.2:1b'
    default_temperature: float = 0.5
    default_output_max_length: int = 500
    default_api_params: dict[str, Any] = {
        'base_url': 'http://localhost:11434/'
    }

    def _setup(self) -> None:
        """Instantiate the Ollama client."""
        self.api_params = copy(
            self.api_params if self.api_params else self.default_api_params
        )
        base_url = self.api_params.pop('base_url')

        self.model = Ollama(
            host=base_url, headers={'x-some-header': 'some-value'}
        )

    def generate(self, query: str, context: list[str]) -> str | BaseModel:
        """
        Generate text by sending a prompt to the local Ollama model.

        Parameters
        ----------
        query : str
            The user query.
        context : list[str]
            Augmented context strings.

        Returns
        -------
        str
            The generated response text.
        """
        input_text = self.prompt_template.format(
            query=query,
            context=' '.join(context),
        )

        messages = []
        if self.system_message:
            messages.append({'role': 'system', 'content': self.system_message})
        messages.append({'role': 'user', 'content': input_text})

        params = {
            'model': self.model_name,
            'messages': messages,
            **(self.api_params or {}),
        }
        response = self.model.chat(**params)
        return str(response.message.content).strip()


@typechecked
class OllamaOpenAIGen(OpenAIGen):
    """OllamaGen via the Ollama Python client."""

    default_model_name = 'llama3.2:1b'
    default_api_params: dict[str, Any] = {
        'base_url': 'http://localhost:11434/v1'
    }

    def _setup(self) -> None:
        self.api_params = copy(
            self.api_params if self.api_params else self.default_api_params
        )
        base_url = self.api_params.pop('base_url')

        model = openai.OpenAI(
            api_key='nokey',
            base_url=base_url,
        )

        self.model = (
            instructor.from_openai(model, mode=instructor.Mode.JSON)
            if self.structured_output
            else model
        )



"""Tests for Rago package: Vector DBs."""

import sys
import tempfile

from functools import partial
from typing import Generator, Optional

import chromadb
import pytest

from chromadb.config import Settings
from rago.augmented.db.chroma import ChromaDB

API_MAP = {
    # ChromaDB: 'api_key_openai',
}

dbs = [
    partial(
        ChromaDB,
        **dict(),
    ),
]


@pytest.fixture
def temp_dir() -> Generator[str, None, None]:
    """Provide temporary directory."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        yield tmp_dir


def create_chroma_client(
    persist_directory: Optional[str] = None,
) -> chromadb.Client:
    """Create a Chroma client instance with specified persist directory."""
    settings = Settings()
    if persist_directory:
        settings = Settings(
            persist_directory=persist_directory, is_persistent=True
        )
    return chromadb.Client(settings=settings)


# def create_chroma_instance(
#     client: chromadb.Client, collection_name: str = 'test_collection'
# ) -> ChromaDB:
#     """Create a Chroma instance with specified client and collection name."""
#     return ChromaDB(client=client, collection_name=collection_name)


@pytest.mark.skipif(
    sys.platform == 'win32',
    reason='Skipping test on Windows due to file locking issues.',
)
@pytest.mark.parametrize(
    'question,expected_answer',
    [
        pytest.param(
            'Is there any animal larger than a dinosaur?', 'Blue Whale', id='0'
        ),
        pytest.param(
            'What animal is renowned as the fastest animal on the planet?',
            'Peregrine Falcon',
            id='1',
        ),
        pytest.param('An animal which do pollination?', 'Honey Bee', id='8'),
    ],
)
@pytest.mark.parametrize('partial_model', dbs)
def test_aug_chromadb(
    request,
    animals_data: list[str],
    question: str,
    expected_answer: str,
    partial_model: partial,
    temp_dir: str,
) -> None:
    """Test RAG pipeline with ChromaDB."""
    question_id = request.node.callspec._idlist[1]
    print(
        f'Running test with ID: {question_id}, \
        question: {question}, \
        expected_answer: {expected_answer}'
    )
    logs = {'augmented': {}}
    top_k = 2
    embedding_size = len(animals_data)  # Choose a fixed embedding size

    # Not needed for this test
    model_class = partial_model.func
    api_key_name: str = API_MAP.get(model_class, '')
    api_key = locals().get(api_key_name, '')

    client = create_chroma_client(temp_dir)

    documents = animals_data
    # Create fixed-size dummy embeddings
    embeddings = [[i] * embedding_size for i in range(len(documents))]

    model_args = {
        'client': client,
        **({'api_key': api_key} if api_key else {}),
    }

    db = partial_model(**model_args)
    db.embed(documents=(documents, embeddings))

    query_encoded = [question_id] * embedding_size
    distances, ids = db.search(query_encoded=query_encoded, top_k=top_k)

    assert len(distances) == 2
    assert len(ids) == 2
    assert ids[0] == question_id
    assert expected_answer.lower() in documents[int(ids[0])].lower()



"""Hugging Face inferencing classes for text generation."""

from __future__ import annotations

from huggingface_hub import InferenceClient
from pydantic import BaseModel
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class HuggingFaceInfGen(GenerationBase):
    """HuggingFaceGen with InferenceClient."""

    default_model_name: str = 'google/gemma-2-2b-it'
    default_api_params = {
        'temperature': 0.7,
        'max_new_tokens': 512,
    }

    def _setup(self) -> None:
        """Set up InferenceClient with API key."""
        self.client = InferenceClient(
            provider='hf-inference', api_key=self.api_key
        )

    def generate(self, query: str, context: list[str]) -> str | BaseModel:
        """Generate the text from the query and augmented context."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )
        if self.system_message:
            input_text = f'{self.system_message}\n{input_text}'

        api_params = self.api_params or self.default_api_params

        self.logs['model_params'] = {
            'model': self.model_name,
            'inputs': input_text,
            'parameters': api_params,
        }
        generated_text = self.client.text_generation(
            prompt=input_text,
            model=self.model_name,
            max_new_tokens=api_params['max_new_tokens'],
            temperature=api_params['temperature'],
        )

        return generated_text.strip()



"""Phi generation module."""

from __future__ import annotations

import warnings

from transformers import AutoModelForCausalLM, AutoTokenizer, GenerationConfig
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class PhiGen(GenerationBase):
    """Phi Generation class."""

    default_model_name: str = 'microsoft/phi-2'
    default_temperature: float = 0.7
    default_output_max_length: int = 500
    default_api_params = {
        'top_p': 0.9,
        'num_return_sequences': 1,
    }

    def _validate(self) -> None:
        """Raise an error if the initial parameters are not valid."""
        if not self.model_name.startswith('microsoft/phi-'):
            raise Exception(
                f'The given model name {self.model_name} is not a Phi model '
                'from Microsoft.'
            )

        if self.structured_output:
            warnings.warn(
                'Structured output is not supported yet in '
                f'{self.__class__.__name__}.'
            )

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        self.tokenizer = AutoTokenizer.from_pretrained(  # type: ignore
            self.model_name, trust_remote_code=True
        )

        device_map = 'auto' if self.device_name == 'cuda' else None

        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_name,
            torch_dtype='auto',
            device_map=device_map,
            trust_remote_code=True,
        )

        if self.tokenizer.pad_token is None:
            self.tokenizer.pad_token = self.tokenizer.eos_token

        self.model.generation_config = GenerationConfig.from_pretrained(
            self.model_name
        )
        self.model.generation_config.pad_token_id = (
            self.model.generation_config.eos_token_id
        )

    def generate(self, query: str, context: list[str]) -> str:
        """Generate text using Phi model with context."""
        full_prompt = f'{query}\nContext: {" ".join(context)}'

        inputs = self.tokenizer(
            full_prompt, return_tensors='pt', return_attention_mask=True
        ).to(self.model.device)

        model_params = dict(
            max_new_tokens=self.output_max_length,
            do_sample=True,
            temperature=self.temperature,
            top_p=self.default_api_params['top_p'],
            num_return_sequences=self.default_api_params[
                'num_return_sequences'
            ],
        )

        self.logs['model_params'] = model_params

        outputs = self.model.generate(
            input_ids=inputs.input_ids,
            attention_mask=inputs.attention_mask,
            **model_params,
        )

        answer: str = self.tokenizer.decode(
            outputs[0][inputs.input_ids.shape[1] :], skip_special_tokens=True
        )

        return answer.strip()



"""TogetherGen class for text generation using Together AI's API."""

from __future__ import annotations

from typing import cast

import instructor
import openai
import together

from pydantic import BaseModel
from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class TogetherGen(GenerationBase):
    """Together AI generation model for text generation."""

    default_model_name: str = 'mistralai/Mixtral-8x7B-Instruct-v0.1'
    default_api_params = {
        'top_p': 0.9,
        'frequency_penalty': 0.0,
        'presence_penalty': 0.0,
    }

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        # if we have to get structured output instructor uses doesn't
        # have support for together yet but we can access the model
        # and use openai sdk if we need to get structured output
        if self.structured_output:
            model = openai.OpenAI(
                base_url='https://api.together.xyz/v1',
                api_key=self.api_key,
            )
            self.model = (
                instructor.from_openai(
                    client=model,
                    mode=instructor.Mode.TOOLS,
                )
                if self.structured_output
                else model
            )

        else:
            together.api_key = self.api_key
            self.model = together.Together()

    def generate(self, query: str, context: list[str]) -> str | BaseModel:
        """Generate text using Together AI's API."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )

        api_params = self.api_params or self.default_api_params

        messages = []
        if self.system_message:
            messages.append({'role': 'system', 'content': self.system_message})
        messages.append({'role': 'user', 'content': input_text})

        model_params = {
            'model': self.model_name,
            'messages': messages,
            'max_tokens': self.output_max_length,
            'temperature': self.temperature,
            **api_params,
        }

        if self.structured_output:
            model_params['response_model'] = self.structured_output
            response = self.model.chat.completions.create(**model_params)
            self.logs['model_params'] = model_params
            return cast(BaseModel, response)

        response = self.model.chat.completions.create(**model_params)
        self.logs['model_params'] = model_params
        return cast(str, response.choices[0].message.content.strip())



"""RAG Generation package."""

from __future__ import annotations

from rago.generation.base import GenerationBase
from rago.generation.cohere import CohereGen
from rago.generation.deepseek import DeepSeekGen
from rago.generation.fireworks import FireworksGen
from rago.generation.gemini import GeminiGen
from rago.generation.groq import GroqGen
from rago.generation.hugging_face import HuggingFaceGen
from rago.generation.hugging_face_inf import HuggingFaceInfGen
from rago.generation.llama import LlamaGen, OllamaGen, OllamaOpenAIGen
from rago.generation.openai import OpenAIGen
from rago.generation.phi import PhiGen
from rago.generation.together import TogetherGen

__all__ = [
    'CohereGen',
    'DeepSeekGen',
    'FireworksGen',
    'GeminiGen',
    'GenerationBase',
    'GroqGen',
    'HuggingFaceGen',
    'HuggingFaceInfGen',
    'LlamaGen',
    'OllamaGen',
    'OllamaOpenAIGen',
    'OpenAIGen',
    'PhiGen',
    'TogetherGen',
]



"""Test the PDF retrieval."""

from pathlib import Path

import pytest

from rago.augmented import SpaCyAug
from rago.retrieval import PDFPathRet

PDF_DATA_PATH = Path(__file__).parent / 'data' / 'pdf'


def test_retrieval_pdf_extraction_basic() -> None:
    """Test the text extraction from a pdf."""
    pdf_ret = PDFPathRet(PDF_DATA_PATH / '1.pdf')
    chunks = pdf_ret.get()

    assert len(chunks) >= 100


@pytest.mark.parametrize(
    'pdf_path,expected',
    [
        ('2407.13797.pdf', ''),
        ('2407.20116.pdf', ''),
    ],
)
def test_retrieval_pdfs_extraction_aug_spacy(
    pdf_path: str, expected: str
) -> None:
    """Test the text extraction from a pdf."""
    pdf_ret = PDFPathRet(PDF_DATA_PATH / pdf_path)
    chunks = pdf_ret.get()

    min_total_chunks = 100  # arbitrary number
    max_chunk_size = pdf_ret.splitter.chunk_size

    assert len(chunks) >= min_total_chunks
    for chunk in chunks:
        assert len(chunk) < max_chunk_size

    query = 'What are the key barriers to implementing vitamin D?'

    aug_top_k = 3

    aug_openai = SpaCyAug(top_k=aug_top_k)
    aug_result = aug_openai.search(query, documents=chunks)

    assert aug_result
    assert len(aug_result) == aug_top_k
    assert all(['vitamin' in result.lower() for result in aug_result])
    assert all(['vitamin d' in result.lower() for result in aug_result])
    assert len(set(aug_result)) == 3
